# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import sys

from _datasketches import TuplePolicy

# This file provides an example Python Tuple Policy implementation.
#
# Each implementation must extend the PyTuplePolicy class and define
# the following methods:
#   * create_summary() returns a new Summary object
#   * update_summary(summary, update) applies the relevant policy to update the
#     provided summary with the data in update.
#   * __call__ may be similar to update_summary but allows a different
#     implementation for set operations (union and intersection)

class AccumulatorPolicy(TuplePolicy):
  '''Implements an accumulatory summary policy, where new values
  are added to the existing value.'''

  def __init__(self):
    TuplePolicy.__init__(self)

  def create_summary(self) -> int:
    return int(0)

  def update_summary(self, summary: int, update: int) -> int:
    summary += update
    return summary

  def __call__(self, summary: int, update: int) -> int:
    summary += update
    return summary


class MaxIntPolicy(TuplePolicy):
  '''Implements a MAX rule, where the largest integer value is always kept.'''

  def __init__(self):
    TuplePolicy.__init__(self)

  def create_summary(self) -> int:
    return int(-sys.maxsize-1)

  def update_summary(self, summary: int, update: int) -> int:
    return max(summary, update)

  def __call__(self, summary: int, update: int) -> int:
    return max(summary, update)


class MinIntPolicy(TuplePolicy):
  '''Implements a MIN rule, where the smallest integer value is always kept.'''

  def __init__(self):
    TuplePolicy.__init__(self)

  def create_summary(self) -> int:
    return int(sys.maxsize)

  def update_summary(self, summary: int, update: int) -> int:
    return min(summary, update)

  def __call__(self, summary: int, update: int) -> int:
    return min(summary, update)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest
from datasketches import cpc_sketch, cpc_union

class CpcTest(unittest.TestCase):
  def test_cpc_example(self):
    lgk = 12    # 2^k = 4096 rows in the table
    n = 1 << 18 # ~256k distinct values

    # create a couple sketches and inject some values
    # we'll have 1/4 of the values overlap
    cpc  = cpc_sketch(lgk)
    cpc2 = cpc_sketch(lgk)
    offset = int(3 * n / 4) # it's a float w/o cast
    # because we hash on the bits, not an abstract numeric value,
    # cpc.update(1) and cpc.update(1.0) give different results.
    for i in range(0, n):
        cpc.update(i)
        cpc2.update(i + offset)

    # although we provide get_composite_estimate() and get_estimate(),
    # the latter will always give the best available estimate.  we
    # recommend using get_estimate().
    # we can check that the upper and lower bounds bracket the
    # estimate, without needing to know the exact value.
    self.assertLessEqual(cpc.get_lower_bound(1), cpc.get_estimate())
    self.assertGreaterEqual(cpc.get_upper_bound(1), cpc.get_estimate())

    # union is a separate class, so we need to get_result()
    # to query the unioned sketches
    union = cpc_union(lgk)
    union.update(cpc)
    union.update(cpc2)
    result = union.get_result()

    # since our process here (including post-union CPC) is
    # deterministic, we have checked and know the exact
    # answer is within one standard deviation of the estimate
    self.assertLessEqual(result.get_lower_bound(1), 7 * n / 4)
    self.assertGreaterEqual(result.get_upper_bound(1), 7 * n / 4)

    # serialize for storage and reconstruct
    sk_bytes = result.serialize()
    new_cpc = cpc_sketch.deserialize(sk_bytes)
    self.assertFalse(new_cpc.is_empty())

    # finally just check that printing works as expected
    self.assertGreater(len(cpc.to_string()), 0)
    self.assertEqual(len(cpc.__str__()), len(cpc.to_string()))

  def test_cpc_get_lg_k(self):
    lgk = 10
    cpc = cpc_sketch(lgk)
    self.assertEqual(cpc.lg_k, lgk)

if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest
from datasketches import tdigest_float, tdigest_double
import numpy as np

class TdigestTest(unittest.TestCase):
    def test_tdigest_double_example(self):
      n = 2 ** 20

      # create a tdigest and inject ~1 million N(0,1) points, both using a vector
      # update as well as a single value
      td = tdigest_double()
      td.update(np.random.normal(size=n-1))
      td.update(0.0)

      # 0 should be near the median
      self.assertAlmostEqual(0.5, td.get_rank(0.0), delta=0.1)

      # the median should be near 0
      self.assertAlmostEqual(0.0, td.get_quantile(0.5), delta=0.1)

      # note that with t-digest, while it typically performs quite well in practice,
      # we do not have any sort of theoretical guarantees on the error bounds
      # or even an estimate of what bounds we may expect.

      # we also track the min/max independently from the rest of the data
      # which lets us know the full observed data range
      self.assertLessEqual(td.get_min_value(), td.get_quantile(0.01))
      self.assertLessEqual(0.0, td.get_rank(td.get_min_value()))
      self.assertGreaterEqual(td.get_max_value(), td.get_quantile(0.99))
      self.assertGreaterEqual(1.0, td.get_rank(td.get_max_value()))

      # and a few basic queries about the sketch
      self.assertFalse(td.is_empty())
      self.assertEqual(td.get_total_weight(), n)

      # we can define a new tdiget with a different distribution, then merge them
      td2 = tdigest_double()
      td2.update(np.random.normal(loc=2.0, size=n))
      td.merge(td2)

      # the new median should be near 1.0, and 1.0 should be near the median although
      # the error distribution is not well-characterized so we allow generous margins
      self.assertAlmostEqual(0.5, td.get_rank(1.0), delta=0.2)
      self.assertAlmostEqual(1.0, td.get_quantile(0.5), delta=0.2)
      self.assertEqual(td.get_total_weight(), 2 * n)

      # finally, can serialize and deserialize the sketch
      td_bytes = td.serialize()
      new_td = tdigest_double.deserialize(td_bytes)
      self.assertEqual(td.get_total_weight(), new_td.get_total_weight())
      self.assertEqual(td.get_min_value(), new_td.get_min_value())
      self.assertEqual(td.get_max_value(), new_td.get_max_value())
      self.assertEqual(td.get_quantile(0.7), new_td.get_quantile(0.7))
      self.assertEqual(td.get_rank(0.0), new_td.get_rank(0.0))


    # the same tests as above, but with tdigest_float
    def test_tdigest_float_example(self):
      n = 2 ** 20
      td = tdigest_float()
      td.update(np.random.normal(size=n-1))
      td.update(0.0)

      self.assertAlmostEqual(0.5, td.get_rank(0.0), delta=0.1)
      self.assertAlmostEqual(0.0, td.get_quantile(0.5), delta=0.1)

      self.assertLessEqual(td.get_min_value(), td.get_quantile(0.01))
      self.assertLessEqual(0.0, td.get_rank(td.get_min_value()))
      self.assertGreaterEqual(td.get_max_value(), td.get_quantile(0.99))
      self.assertGreaterEqual(1.0, td.get_rank(td.get_max_value()))

      self.assertFalse(td.is_empty())
      self.assertEqual(td.get_total_weight(), n)

      td2 = tdigest_float()
      td2.update(np.random.normal(loc=2.0, size=n))
      td.merge(td2)

      self.assertAlmostEqual(0.5, td.get_rank(1.0), delta=0.2)
      self.assertAlmostEqual(1.0, td.get_quantile(0.5), delta=0.2)
      self.assertEqual(td.get_total_weight(), 2 * n)

      td_bytes = td.serialize()
      new_td = tdigest_float.deserialize(td_bytes)
      self.assertEqual(td.get_total_weight(), new_td.get_total_weight())
      self.assertEqual(td.get_min_value(), new_td.get_min_value())
      self.assertEqual(td.get_max_value(), new_td.get_max_value())
      self.assertEqual(td.get_quantile(0.7), new_td.get_quantile(0.7))
      self.assertEqual(td.get_rank(0.0), new_td.get_rank(0.0))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
  
import unittest
from datasketches import count_min_sketch

class CountMinTest(unittest.TestCase):
  def test_count_min_example(self):
    # we'll define target confidence and relative error and use the built-in
    # methods to determine how many hashes and buckets to use
    confidence = 0.95
    num_hashes = count_min_sketch.suggest_num_hashes(confidence)
    relative_error = 0.01
    num_buckets = count_min_sketch.suggest_num_buckets(relative_error)

    # now we can create a few empty sketches
    cm = count_min_sketch(num_hashes, num_buckets)
    cm2 = count_min_sketch(num_hashes, num_buckets)
    self.assertTrue(cm.is_empty())

    # we'll use a moderate number of distinct items with
    # increasing weights, with each item's weight being
    # equal to its value
    n = 1000
    total_wt = 0
    for i in range(1, n+1):
      cm.update(i, i)
      total_wt += i
    self.assertFalse(cm.is_empty())
    self.assertEqual(cm.total_weight, total_wt)

    # querying the items, each of them should
    # have a non-zero count.  the estimate should
    # be at least i with appropriately behaved bounds.
    for i in range(1, n+1):
      val = cm.get_estimate(i)
      self.assertGreaterEqual(val, i)
      self.assertGreaterEqual(val, cm.get_lower_bound(i))
      self.assertGreater(cm.get_upper_bound(i), val)

    # values not in the sketch should have lower estimates, but
    # are not guaranteed to be zero and will succeed
    self.assertIsNotNone(cm.get_estimate("not in set"))

    # we can create another sketch with partial overlap
    # and merge them
    for i in range(int(n / 2), int(3 * n / 2)):
      cm2.update(i, i)
    cm.merge(cm2)

    # and the estimated weight for the overlapped meerged values
    # (n/2 to n) should now be at least 2x the value
    self.assertGreaterEqual(cm.get_estimate(n), 2 * n)

    # finally, serialize and reconstruct
    cm_bytes = cm.serialize()
    self.assertEqual(cm.get_serialized_size_bytes(), len(cm_bytes))
    new_cm = count_min_sketch.deserialize(cm_bytes)

    # and now interrogate the sketch
    self.assertFalse(new_cm.is_empty())
    self.assertEqual(new_cm.num_hashes, cm.num_hashes)
    self.assertEqual(new_cm.num_buckets, cm.num_buckets)
    self.assertEqual(new_cm.total_weight, cm.total_weight)
    
    # we can also iterate through values in and out of the sketch to ensure
    # the estimates match
    for i in range(0, 2 * n):
      self.assertEqual(cm.get_estimate(i), new_cm.get_estimate(i))

    # finally just check that printing works as expected
    self.assertGreater(len(cm.to_string()), 0)
    self.assertEqual(len(cm.__str__()), len(cm.to_string()))

if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import copy
import unittest
from datasketches import quantiles_ints_sketch, quantiles_floats_sketch, quantiles_doubles_sketch
from datasketches import quantiles_items_sketch, ks_test, PyStringsSerDe
import numpy as np

class QuantilesTest(unittest.TestCase):
    def test_quantiles_floats_example(self):
      k = 128
      n = 2 ** 20

      # create a sketch and inject ~1 million N(0,1) points as an array and as a single item
      quantiles = quantiles_floats_sketch(k)
      quantiles.update(np.random.normal(size=n-1))
      quantiles.update(0.0)

      # 0 should be near the median
      self.assertAlmostEqual(0.5, quantiles.get_rank(0.0), delta=0.035)
      
      # the median should be near 0
      self.assertAlmostEqual(0.0, quantiles.get_quantile(0.5), delta=0.035)

      # we also track the min/max independently from the rest of the data
      # which lets us know the full observed data range
      self.assertLessEqual(quantiles.get_min_value(), quantiles.get_quantile(0.01))
      self.assertLessEqual(0.0, quantiles.get_rank(quantiles.get_min_value()))
      self.assertGreaterEqual(quantiles.get_max_value(), quantiles.get_quantile(0.99))
      self.assertGreaterEqual(1.0, quantiles.get_rank(quantiles.get_max_value()))

      # we can also extract a list of values at a time,
      # here the values should give us something close to [-2, -1, 0, 1, 2].
      # then get the CDF, which will return something close to
      # the original values used in get_quantiles()
      # finally, can check the normalized rank error bound
      pts = quantiles.get_quantiles([0.0228, 0.1587, 0.5, 0.8413, 0.9772])
      cdf = quantiles.get_cdf(pts)  # include 1.0 at end to account for all probability mass
      self.assertEqual(len(cdf), len(pts)+1)
      err = quantiles.normalized_rank_error(False)
      self.assertEqual(err, quantiles_floats_sketch.get_normalized_rank_error(k, False))

      # and a few basic queries about the sketch
      self.assertFalse(quantiles.is_empty())
      self.assertTrue(quantiles.is_estimation_mode())
      self.assertEqual(quantiles.n, n)
      self.assertEqual(quantiles.k, k)
      self.assertLess(quantiles.num_retained, n)

      # merging itself will double the number of items the sketch has seen
      quantiles_copy = copy.copy(quantiles)
      quantiles.merge(quantiles_copy)
      self.assertEqual(quantiles.n, 2*n)

      # we can then serialize and reconstruct the sketch
      quantiles_bytes = quantiles.serialize()
      new_quantiles = quantiles_floats_sketch.deserialize(quantiles_bytes)
      self.assertEqual(quantiles.num_retained, new_quantiles.num_retained)
      self.assertEqual(quantiles.get_min_value(), new_quantiles.get_min_value())
      self.assertEqual(quantiles.get_max_value(), new_quantiles.get_max_value())
      self.assertEqual(quantiles.get_quantile(0.7), new_quantiles.get_quantile(0.7))
      self.assertEqual(quantiles.get_rank(0.0), new_quantiles.get_rank(0.0))

      # If we create a new sketch with a very different distribution, a Kolmogorov-Smirnov Test
      # of the two should return True: we can reject the null hypothesis that the sketches
      # come from the same distributions.
      unif_quantiles = quantiles_floats_sketch(k)
      unif_quantiles.update(np.random.uniform(10, 20, size=n-1))
      self.assertTrue(ks_test(quantiles, unif_quantiles, 0.001))

      total_weight = 0
      for tuple in quantiles:
        item = tuple[0]
        weight = tuple[1]
        total_weight = total_weight + weight
      self.assertEqual(total_weight, quantiles.n)

    def test_quantiles_ints_sketch(self):
        k = 128
        n = 10
        quantiles = quantiles_ints_sketch(k)
        for i in range(0, n):
          quantiles.update(i)

        self.assertEqual(quantiles.get_min_value(), 0)
        self.assertEqual(quantiles.get_max_value(), n-1)
        self.assertEqual(quantiles.n, n)
        self.assertFalse(quantiles.is_empty())
        self.assertFalse(quantiles.is_estimation_mode()) # n < k
        self.assertEqual(quantiles.k, k)

        pmf = quantiles.get_pmf([round(n/2)])
        self.assertIsNotNone(pmf)
        self.assertEqual(len(pmf), 2)

        cdf = quantiles.get_cdf([round(n/2)])
        self.assertIsNotNone(cdf)
        self.assertEqual(len(cdf), 2)

        self.assertEqual(quantiles.get_quantile(0.5), round(n/2))
        quants = quantiles.get_quantiles([0.25, 0.5, 0.75])
        self.assertIsNotNone(quants)
        self.assertEqual(len(quants), 3)

        self.assertEqual(quantiles.get_rank(round(n/2)), 0.5)

        # merge self
        quantiles_copy = copy.copy(quantiles)
        quantiles.merge(quantiles_copy)
        self.assertEqual(quantiles.n, 2 * n)

        sk_bytes = quantiles.serialize()
        self.assertTrue(isinstance(quantiles_ints_sketch.deserialize(sk_bytes), quantiles_ints_sketch))

    def test_quantiles_doubles_sketch(self):
      # already tested floats and ints and it's templatized, so just make sure it instantiates properly
      k = 128
      quantiles = quantiles_doubles_sketch(k)
      self.assertTrue(quantiles.is_empty())

    def test_quantiles_items_sketch(self):
      # most functionality has been tested, but we need to ensure objects and sorting work
      # as well as serialization
      k = 128
      n = 2 ** 16

      # create a sketch and inject enough points to force compaction
      quantiles = quantiles_items_sketch(k)
      for i in range(0, n):
        quantiles.update(str(i))
      
      quantiles_copy = copy.copy(quantiles)
      quantiles.merge(quantiles_copy)
      self.assertEqual(quantiles.n, 2 * n)
      
      total_weight = 0
      for tuple in quantiles:
        item = tuple[0]
        weight = tuple[1]
        total_weight = total_weight + weight
      self.assertEqual(total_weight, quantiles.n)

      quantiles_bytes = quantiles.serialize(PyStringsSerDe())
      new_quantiles = quantiles_items_sketch.deserialize(quantiles_bytes, PyStringsSerDe())
      self.assertEqual(quantiles.num_retained, new_quantiles.num_retained)
      self.assertEqual(quantiles.get_min_value(), new_quantiles.get_min_value())
      self.assertEqual(quantiles.get_max_value(), new_quantiles.get_max_value())
      self.assertEqual(quantiles.get_quantile(0.7), new_quantiles.get_quantile(0.7))
      self.assertEqual(quantiles.get_rank(str(n/4)), new_quantiles.get_rank(str(n/4)))

      # The sketches are identical so we cannot reject the null hypothesis that they
      # reflect the same underlying distribtion
      self.assertFalse(ks_test(quantiles, new_quantiles, 0.001))

      # check that printing works as expected
      self.assertGreater(len(quantiles.to_string(True, True)), 0)
      self.assertEqual(len(quantiles.__str__()), len(quantiles.to_string()))

if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from _datasketches import PyObjectSerDe

import struct

# This file provides several Python SerDe implementation examples.
#
# Each implementation must extend the PyObjectSerDe class and define
# three methods:
#   * get_size(item) returns an int of the number of bytes needed to
#     serialize the given item
#   * to_bytes(item) returns a bytes object representing a serialized
#     version of the given item
#   * from_bytes(data, offset) takes a bytes object (data) and an offset
#     indicating where in the data array to start reading. The method
#     returns a tuple with the newly reconstructed object and the
#     total number of bytes beyond the offset read from the input data.

class PyStringsSerDe(PyObjectSerDe):
  '''Implements a simple string-encoding scheme where a string is
     written as `<num_bytes> <string_contents>`, with no null termination.
     This format allows pre-allocating each string, at the cost of
     additional storage. Using this format, the serialized string consumes
     ``4 + len(item)`` bytes.'''
  def get_size(self, item):
    return int(4 + len(item))

  def to_bytes(self, item: str):
    b = bytearray()
    b.extend(len(item).to_bytes(4, 'little'))
    b.extend(map(ord,item))
    return bytes(b)

  def from_bytes(self, data: bytes, offset: int):
    num_chars = int.from_bytes(data[offset:offset+3], 'little')
    if (num_chars < 0 or num_chars > offset + len(data)):
        raise IndexError(f'num_chars read must be non-negative and not larger than the buffer. Found {num_chars}')
    str = data[offset+4:offset+4+num_chars].decode()
    return (str, 4+num_chars)

class PyIntsSerDe(PyObjectSerDe):
  '''Implements an integer encoding scheme where each integer is written
     as a 32-bit (4 byte) little-endian value.'''
  def get_size(self, item):
    return int(4)

  def to_bytes(self, item):
    return struct.pack('<i', item)

  def from_bytes(self, data: bytes, offset: int):
    val = struct.unpack_from('<i', data, offset)[0]
    return (val, 4)


class PyLongsSerDe(PyObjectSerDe):
  '''Implements an integer encoding scheme where each integer is written
     as a 64-bit (8 byte) little-endian value.'''
  def get_size(self, item):
    return int(8)

  def to_bytes(self, item):
    return struct.pack('<q', item)

  def from_bytes(self, data: bytes, offset: int):
    val = struct.unpack_from('<q', data, offset)[0]
    return (val, 8)


class PyFloatsSerDe(PyObjectSerDe):
  '''Implements a floating point encoding scheme where each value is written
     as a 32-bit floating point value.'''
  def get_size(self, item):
    return int(4)

  def to_bytes(self, item):
    return struct.pack('<f', item)

  def from_bytes(self, data: bytes, offset: int):
    val = struct.unpack_from('<f', data, offset)[0]
    return (val, 4)


class PyDoublesSerDe(PyObjectSerDe):
  '''Implements a floating point encoding scheme where each value is written
     as a 64-bit floating point value.'''
  def get_size(self, item):
    return int(8)

  def to_bytes(self, item):
    return struct.pack('<d', item)

  def from_bytes(self, data: bytes, offset: int):
    val = struct.unpack_from('<d', data, offset)[0]
    return (val, 8)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import numpy as np

from _datasketches import KernelFunction

# This file provides an example Python Kernel Function implementation.
#
# Each implementation must extend the KernelFunction class
# and define the __call__ method

class GaussianKernel(KernelFunction):
  '''Implements a basic Gaussian kernel

     :param bandwidth: The kernel bandwidth, default 1.0
     :type bandwidth: float
    '''
  def __init__(self, bandwidth: float=1.0):
    KernelFunction.__init__(self)
    self._bw = bandwidth
    self._scale = -0.5 * (bandwidth ** -2)

  def __call__(self, a: np.ndarray, b: np.ndarray) -> float:
    return np.exp(self._scale * np.linalg.norm(a - b)**2)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest
from datasketches import density_sketch, KernelFunction, GaussianKernel
import numpy as np

class UnitSphereKernel(KernelFunction):
  def __call__(self, a: np.ndarray, b: np.ndarray) -> float:
    if np.linalg.norm(a - b) < 1.0:
      return 1.0
    else:
      return 0.0

class densityTest(unittest.TestCase):
  def test_density_sketch(self):
    k = 10
    dim = 3
    n = 1000

    sketch = density_sketch(k, dim, GaussianKernel())

    self.assertEqual(sketch.k, k)
    self.assertEqual(sketch.dim, dim)
    self.assertTrue(sketch.is_empty())
    self.assertFalse(sketch.is_estimation_mode())
    self.assertEqual(sketch.n, 0)
    self.assertEqual(sketch.num_retained, 0)

    for i in range(n):
      sketch.update([i, i, i])

    self.assertFalse(sketch.is_empty())
    self.assertTrue(sketch.is_estimation_mode())
    self.assertEqual(sketch.n, n)
    self.assertGreater(sketch.num_retained, k)
    self.assertLess(sketch.num_retained, n)
    self.assertGreater(sketch.get_estimate([n - 1, n - 1, n - 1]), 0)

    for tuple in sketch:
      vector = tuple[0]
      weight = tuple[1]
      self.assertEqual(len(vector), dim)
      self.assertGreaterEqual(weight, 1)

    sk_bytes = sketch.serialize()
    sketch2 = density_sketch.deserialize(sk_bytes, GaussianKernel())
    self.assertEqual(sketch.get_estimate([1.5, 2.5, 3.5]), sketch2.get_estimate([1.5, 2.5, 3.5]))

    # check that printing works as expected
    self.assertGreater(len(sketch.to_string(True, True)), 0)
    self.assertEqual(len(sketch.__str__()), len(sketch.to_string()))

  def test_density_merge(self):
    sketch1 = density_sketch(10, 2, GaussianKernel())
    sketch1.update([0, 0])
    sketch2 = density_sketch(10, 2, GaussianKernel())
    sketch2.update([0, 1])
    sketch1.merge(sketch2)
    self.assertEqual(sketch1.n, 2)
    self.assertEqual(sketch1.num_retained, 2)

  def test_custom_kernel(self):
    gaussianSketch = density_sketch(10, 2, GaussianKernel())
    sphericalSketch = density_sketch(10, 2, UnitSphereKernel())

    p = [1, 1]
    gaussianSketch.update(p)
    sphericalSketch.update(p)

    # Spherical kernel should return 1.0 for a nearby point, 0 farther
    # Gaussian kernel should return something nonzero when farther away
    self.assertEqual(sphericalSketch.get_estimate([1.001, 1]), 1.0)
    self.assertEqual(sphericalSketch.get_estimate([2, 2]), 0.0)
    self.assertGreater(gaussianSketch.get_estimate([2, 2]), 0.0)

    # We can also use a custom kernel when deserializing
    sk_bytes = sphericalSketch.serialize()
    sphericalRebuilt = density_sketch.deserialize(sk_bytes, UnitSphereKernel())
    self.assertEqual(sphericalSketch.get_estimate([1.001, 1]), sphericalRebuilt.get_estimate([1.001, 1]))

if __name__ == '__main__':
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# Modified from:
# http://www.benjack.io/2018/02/02/python-cpp-revisited.html

import os
import sys
import platform
import subprocess
import re
from datetime import datetime, timezone

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    def run(self):
        try:
            subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake >= 3.12 must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args =  ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir]
        # ensure we use a consistent python version
        #cmake_args += ['-DPython3_EXECUTABLE=' + sys.executable]
        cmake_args += ['-DPython_EXECUTABLE=' + sys.executable]
        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(
                           cfg.upper(),
                           extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-T', 'host=x64']
                cmake_args += ['-DCMAKE_GENERATOR_PLATFORM=x64']
                build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get('CXXFLAGS', ''),
            self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args,
                              cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.', '--target', 'python'] + build_args,
                              cwd=self.build_temp, env=env)
        print() # add an empty line to pretty print

# Read and parse the version format
# @DT@ -> datestamp
# @HHMM@ -> .devHHMM to indicate development version
# Releases should have a fixed version with no @ variables
with open('version.cfg.in', 'r') as file:
    ds_version = file.read().rstrip()
dt = datetime.now(timezone.utc)
ds_version = re.sub('@DT@', dt.strftime('%Y%m%d'), ds_version)
ds_version = re.sub('@HHMM@', 'dev' + dt.strftime('%H%M'), ds_version)

setup(
    name='datasketches',
    version=ds_version,
    author='Apache Software Foundation',
    author_email='dev@datasketches.apache.org',
    description='The Apache DataSketches Library for Python',
    license='Apache License 2.0',
    url='http://datasketches.apache.org',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['src','include','*tests*']), # src not needed if only the .so
    package_dir={'':'.'},
    # may need to add all source paths for sdist packages w/o MANIFEST.in
    ext_modules=[CMakeExtension('datasketches','.')],
    cmdclass={'build_ext': CMakeBuild},
    install_requires=['numpy'],
    zip_safe=False
)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
 
import unittest
from math import floor, ceil
from datasketches import ebpps_sketch, PyIntsSerDe, PyStringsSerDe

class EbppsTest(unittest.TestCase):
  def test_ebpps_example(self):
    k = 50  # a small value so we can easily fill the sketch
    sk = ebpps_sketch(k)

    # ebpps sampling reduces to standard reservoir sampling
    # if the items are all equally weighted, although the
    # algorithm will be significantly slower than an optimized
    # reservoir sampler
    n = 5 * k
    for i in range(0, n):
      sk.update(i)
    
    # we should have k samples since they're equally weighted
    self.assertAlmostEqual(k, sk.c, places=12)

    # we can also add a heavy item, using a negative value to
    # be able to identify the item later.  Keep in mind that
    # "heavy" is a relative concept, so using a fixed
    # multiple of n may not be considered a heavy item for
    # larger values of n
    sk.update(-1, 1000 * n)
    self.assertEqual(k, sk.k)
    self.assertLess(sk.c, k)
    self.assertEqual(n + 1, sk.n)
    self.assertFalse(sk.is_empty())

    # we can easily get the list of items in the sample
    items = list(sk)
    self.assertLessEqual(len(items), k)

    count = 0
    for items in sk:
      count = count + 1
    self.assertTrue(count == floor(sk.c) or count == ceil(sk.c))

    # next we'll create a second, smaller sketch with
    # only heavier items relative to the previous sketch
    k2 = 5
    sk2 = ebpps_sketch(k2)
    # for weight, use the sum of all items >=0 from before (n)
    wt = n
    for i in range(0, k2 + 1):
      sk2.update((2 * n) + i, wt)

    # now merge the sketches, noting how the
    # resulting k will be the smaller of the two.
    input_sk_c = sk.c
    sk.merge(sk2)

    self.assertEqual(n + k2 + 2, sk.n)
    self.assertFalse(sk.is_empty())
    self.assertEqual(sk.k, k2)
    self.assertLess(sk.k, k)

    # the expected number of samples post-merge may be larger than
    # with the input sketch
    self.assertGreater(sk.c, input_sk_c)

    # we can dump a sumamry of information in the sketch
    print(sk)

    # if we want to print the list of items, there must be a
    # __str__() method for each item (which need not be the same
    # type; they're all generic python objects when used from
    # python), otherwise you may trigger an exception.
    # to_string() is provided as a convenience to avoid direct
    # calls to __str__() with parameters.
    print(sk.to_string(True))

    # finally, we can serialize the sketch by providing an
    # appropriate serde class.
    expected_size = sk.get_serialized_size_bytes(PyIntsSerDe())
    b = sk.serialize(PyIntsSerDe())
    self.assertEqual(expected_size, len(b))

    # if we try to deserialize with the wrong serde, things break
    try:
      ebpps_sketch.deserialize(b, PyStringsSerDe())
      self.fail()
    except:
      # expected; do nothing
      self.assertTrue(True)
      
    # using the correct serde gives us back a copy of the original
    rebuilt = ebpps_sketch.deserialize(b, PyIntsSerDe())
    self.assertEqual(sk.k, rebuilt.k)
    self.assertEqual(sk.c, rebuilt.c)
    self.assertEqual(sk.n, rebuilt.n)

    # check that printing works as expected
    self.assertGreater(len(sk.to_string(True)), 0)
    self.assertEqual(len(sk.__str__()), len(sk.to_string()))

if __name__ == '__main__':
  unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import unittest

from datasketches import update_tuple_sketch
from datasketches import compact_tuple_sketch, tuple_union
from datasketches import tuple_intersection, tuple_a_not_b
from datasketches import tuple_jaccard_similarity
from datasketches import tuple_jaccard_similarity, PyIntsSerDe
from datasketches import AccumulatorPolicy, MaxIntPolicy, MinIntPolicy
from datasketches import update_theta_sketch

class TupleTest(unittest.TestCase):
    def test_tuple_basic_example(self):
        lgk = 12    # 2^k = 4096 rows in the table
        n = 1 << 18 # ~256k unique values

        # create a sketch and inject some values -- summary is 2 so we can sum them
        # and know the result
        sk = self.generate_tuple_sketch(AccumulatorPolicy(), n, lgk, value=2)

        # we can check that the upper and lower bounds bracket the
        # estimate, without needing to know the exact value.
        self.assertLessEqual(sk.get_lower_bound(1), sk.get_estimate())
        self.assertGreaterEqual(sk.get_upper_bound(1), sk.get_estimate())

        # because this sketch is deterministically generated, we can
        # also compare against the exact value
        self.assertLessEqual(sk.get_lower_bound(1), n)
        self.assertGreaterEqual(sk.get_upper_bound(1), n)

        # compact and serialize for storage, then reconstruct
        sk_bytes = sk.compact().serialize(PyIntsSerDe())
        new_sk = compact_tuple_sketch.deserialize(sk_bytes, serde=PyIntsSerDe())

        # estimate remains unchanged
        self.assertFalse(sk.is_empty())
        self.assertEqual(sk.get_estimate(), new_sk.get_estimate())

        # we can also iterate over the sketch entries
        # the iterator provides a (hashkey, summary) pair where the
        # first value is the raw hash value and the second the summary
        count = 0
        cumSum = 0
        for pair in new_sk:
          self.assertLess(pair[0], new_sk.theta64)
          count += 1
          cumSum += pair[1]
        self.assertEqual(count, new_sk.num_retained)
        self.assertEqual(cumSum, 2 * new_sk.num_retained)

        # we can even create a tuple sketch from an existing theta sketch
        # as long as we provide a summary to use
        theta_sk = update_theta_sketch(lgk)
        for i in range(n, 2*n):
          theta_sk.update(i)
        cts = compact_tuple_sketch(theta_sk, 5)
        cumSum = 0
        for pair in cts:
          cumSum += pair[1]
        self.assertEqual(cumSum, 5 * cts.num_retained)

        # check that printing works as expected
        self.assertGreater(len(sk.to_string(True)), 0)
        self.assertEqual(len(sk.__str__()), len(sk.to_string()))

        num = sk.num_retained
        sk.trim()
        self.assertLessEqual(sk.num_retained, num)

        sk.reset()
        self.assertTrue(sk.is_empty())
        self.assertEqual(sk.num_retained, 0)

    def test_tuple_filter(self):
        lgk = 12    # 2^k = 4096 rows in the table
        n = 1 << 18 # ~256k unique values

        # filtering lets us apply a predicate to the sketch, producing a new
        # compact sketch using the entries matching the predicate.
        sk = update_tuple_sketch(AccumulatorPolicy(), lgk)
        for ii in range(0, n):
          sk.update(ii, ii)

        # we can filter by a predicate, whether a lambda or a defined function
        # for instance, using 0.5*n will return a compact_tuple_sketch with
        # approximately half the entries.
        result = sk.filter(lambda x: x < (0.5 * n))
        self.assertAlmostEqual(result.get_estimate(), 0.5 * n, delta=0.01 * n)
        self.assertLess(result.get_lower_bound(1), 0.5 * n)
        self.assertGreater(result.get_upper_bound(1), 0.5 * n)

    def test_tuple_set_operations(self):
        lgk = 12    # 2^k = 4096 rows in the table
        n = 1 << 18 # ~256k unique values

        # we'll have 1/4 of the values overlap
        offset = int(3 * n / 4) # it's a float w/o cast

        # create a couple sketches and inject some values, with different summaries
        sk1 = self.generate_tuple_sketch(AccumulatorPolicy(), n, lgk, value=5)
        sk2 = self.generate_tuple_sketch(AccumulatorPolicy(), n, lgk, value=7, offset=offset)

        # UNIONS
        # create a union object
        union = tuple_union(MaxIntPolicy(), lgk)
        union.update(sk1)
        union.update(sk2)

        # getting result from union returns a compact_theta_sketch
        # compact theta sketches can be used in additional unions
        # or set operations but cannot accept further item updates
        result = union.get_result()
        self.assertTrue(isinstance(result, compact_tuple_sketch))

        # since our process here is deterministic, we have
        # checked and know the exact answer is within one
        # standard deviation of the estimate
        self.assertLessEqual(result.get_lower_bound(1), 7 * n / 4)
        self.assertGreaterEqual(result.get_upper_bound(1), 7 * n / 4)

        # we unioned two equal-sized sketches with overlap and used
        # the max value as the resulting summary, meaning we should
        # have more summaries with value 7 than value 5 in the result
        count5 = 0
        count7 = 0
        for pair in result:
          if pair[1] == 5:
            count5 += 1
          elif pair[1] == 7:
            count7 += 1
          else:
            self.fail()
        self.assertLess(count5, count7)

        # INTERSECTIONS
        # create an intersection object
        intersect = tuple_intersection(MinIntPolicy()) # no lg_k
        intersect.update(sk1)
        intersect.update(sk2)

        # has_result() indicates the intersection has been used,
        # although the result may be the empty set
        self.assertTrue(intersect.has_result())

        # as with unions, the result is a compact sketch
        result = intersect.get_result()
        self.assertTrue(isinstance(result, compact_tuple_sketch))

        # we know the sets overlap by 1/4
        self.assertLessEqual(result.get_lower_bound(1), n / 4)
        self.assertGreaterEqual(result.get_upper_bound(1), n / 4)

        # in this example, we intersected the sketches and took the
        # min value as the resulting summary, so all summaries
        # must be exactly equal to that value
        count5 = 0
        for pair in result:
          if pair[1] == 5:
            count5 += 1
          else:
            self.fail()
        self.assertEqual(count5, result.num_retained)

        # A NOT B
        # create an a_not_b object
        anb = tuple_a_not_b() # no lg_k or policy
        result = anb.compute(sk1, sk2)

        # as with unions, the result is a compact sketch
        self.assertTrue(isinstance(result, compact_tuple_sketch))

        # we know the sets overlap by 1/4, so the remainder is 3/4
        self.assertLessEqual(result.get_lower_bound(1), 3 * n / 4)
        self.assertGreaterEqual(result.get_upper_bound(1), 3 * n / 4)

        # here, we have only values with a summary of 5 as any keys that
        # existed in both sketches were removed
        count5 = 0
        for pair in result:
          if pair[1] == 5:
            count5 += 1
          else:
            self.fail()
        self.assertEqual(count5, result.num_retained)

        # JACCARD SIMILARITY
        # Jaccard Similarity measure returns (lower_bound, estimate, upper_bound)
        # and does not examine summaries, even for (dis)similarity tests.
        jac = tuple_jaccard_similarity.jaccard(sk1, sk2)

        # we can check that results are in the expected order
        self.assertLess(jac[0], jac[1])
        self.assertLess(jac[1], jac[2])

        # checks for sketch equivalence
        self.assertTrue(tuple_jaccard_similarity.exactly_equal(sk1, sk1))
        self.assertFalse(tuple_jaccard_similarity.exactly_equal(sk1, sk2))

        # we can apply a check for similarity or dissimilarity at a
        # given threshold, at 97.7% confidence.

        # check that the Jaccard Index is at most (upper bound) 0.2.
        # exact result would be 1/7
        self.assertTrue(tuple_jaccard_similarity.dissimilarity_test(sk1, sk2, 0.2))

        # check that the Jaccard Index is at least (lower bound) 0.7
        # exact result would be 3/4, using result from A NOT B test
        self.assertTrue(tuple_jaccard_similarity.similarity_test(sk1, result, 0.7))


    # Generates a basic tuple sketch with a fixed value for each update
    def generate_tuple_sketch(self, policy, n, lgk, value, offset=0):
      sk = update_tuple_sketch(policy, lgk)
      for i in range(0, n):
        sk.update(i + offset, value)
      return sk

if __name__ == '__main__':
    unittest.main()
